/********************************************************************
*
*
*
*********************************************************************
* FileName:        Touch.c
* Dependencies:    See INCLUDES section below
* Processor:       PIC16F886
* Compiler:        HITECH 9.60 STD (only)
* Company:         Microchip Technology, Inc.
*
* Software License Agreement:
*
* The software supplied herewith by Microchip Technology Incorporated
* (the "Company") for its PICmicro Microcontroller is intended and
* supplied to you, the Company's customer, for use solely and
* exclusively on Microchip PICmicro Microcontroller products. The
* software is owned by the Company and/or its supplier, and is
* protected under applicable copyright laws. All rights are reserved.
* Any use in violation of the foregoing restrictions may subject the
* user to criminal sanctions under applicable laws, as well as to
* civil liability for the breach of the terms and conditions of this
* license.
*
* THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION. NO WARRANTIES,
* WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
* TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
* PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
* IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
* CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
*********************************************************************
* File Description: Main touch program
*
* Change History:
* Author k.curtis   Date 2008       Comment
********************************************************************/

#include "pic.h"

__CONFIG(INTIO & WDTDIS & PWRTDIS & MCLREN & UNPROTECT & BORDIS & IESODIS & FCMDIS & LVPDIS & DEBUGDIS & BORV40);

#include "define.h"
#include "ssp.h"

unsigned int delay;
unsigned char padcount;
unsigned char anykeypressed;
unsigned char padpressed;
unsigned int downkeys;
unsigned char x;
signed long largetemp;								// Single large variable for calculations

bank1 signed int both_coils;						// conversion value for both coils
bank1 signed int ref_coil;							// conversion value for just the reference coil
bank1 signed int gnd_ref;							// conversion value for the virtual ground
bank1 signed int threshold;							// conversion value for the press threshold
bank1 signed int norm_btn;							// conversion value for the normalized button value

unsigned int keypressed;
unsigned char testbit;

// I2C Communications
unsigned char	ByteCount;
unsigned int	TempInt;
unsigned char	SSPCommand;
unsigned char	SSPIndex;
unsigned char	SSPSize;
unsigned char	IndexCount;

FFlags Flags;

bank3 unsigned int	RAW	   [NUM_BTTNS];				// Raw data from converter
bank3 unsigned int	AVERAGE[NUM_BTTNS];				// running average
bank3 unsigned int	TRIP   [NUM_BTTNS];				// holds threshold offset for press detect
bank3 unsigned int	GUARD  [NUM_BTTNS];				// not used, retained for GUI compatability

unsigned int	value;								// current button value

void interrupt ISR(void)
{
 	// SSP Interrupt
	if (SSPIF == 1 && SSPIE == 1)					// SSP Interrupt Detected
		SSPISR();
}


void main(void)
{

	unsigned char led;

	OSCCON = 0x71;	 				// Internal oscillator 8mhz
	OSCTUNE= 0x0	;				// Oscillator trim - Microchip data sheet suspect here!
									// - should default to this but doesn't always seem to!
	while ((OSCCON & 0x04) == 0);	// wait for clock to stabilise

	InitPORT();

	for (led = 0; led < NUM_BTTNS; led++)
		{
			SetLED(led);
			restart_delay();		// Settle down delay
			restart_delay();		// Settle down delay
			restart_delay();		// Settle down delay
			restart_delay();		// Settle down delay
			SetLED(NUM_BTTNS);
			restart_delay();		// Settle down delay
			restart_delay();		// Settle down delay
		}

	InitPeripheral();

	keypressed = 0;					// Initialise a few variables
	restart_delay();				// Settle down delay
	downkeys = 0;
	padcount = 0;
	// ********* INTERRUPT SETUP ************

	GIE = 1;						// enable global interrupts
	PEIE = 1;						// enable peripheral interrupts
									// .. these are INTCON register bits
/*
Thresholds and Averaging
=====================
Because there is bound to be variation in the coil gaps between keys, and between units due to the method of construction used,
the method of setting the threshold of operation for each key is as follows.
An array is generated which contains the threshold for each key. This is held as a 16 times multiple of the key reading value.
Each time a keypad is read the returned value (assuming the key is not pressed) is added to the corresponding array value,
and the result is multiplied by 16 using an arithmetic shift (for speed). This figure is then divided by 17 and returned to the array.
The net effect of this will be no change if the returned value is constant. If the value drifts with temperature or any other factor
the array value will slowly track the change. This method give good noise immunity without excessive RAM or calculation
overhead, as only one 16 bit value is stored per key, and the arithmetic is minimal, consisting mainly of one divide operation per cycle.
*/

/* Here we initialise the threshold array to 16 times the key reading less an offset.
For practical reasons it is better to err towards a low reading which will self-adjust as the value tracks,
so we subtract the nominal threshold value from each reading to ensure that the start point is too low, rather than too high.
Too high an initial value could cause a situation where a latch up occurs as the software would consider the key to be pressed,
and not therefore change its average value. This method provides a good compromise.
*/
	x=0;
	do													// for each of 4 keys
	{

		measure_pad(x);									// get the returned value for pad
		AVERAGE[x] = norm_btn;							// load into array
		AVERAGE[x] <<= 4;								// multiply by 16
		TRIP[x] = AVERAGE[x] * SENSIBILITY;
		x++;
	}
	while (x<NUM_BTTNS);


	while (1)											// Main program loop
	{
		if (padcount == 0)
		{
			anykeypressed = 0;							// reset anykeypressed flag at the start of each cycle
			SetLED(NUM_BTTNS);							// Turn off LED
		}
		measure_pad(padcount);							// Do actual pad read
		if (padpressed == 1)
		{
			testbit  = ( keypressed >> padcount)& 0x01;
			if (testbit == 0)
				// This detects the unpressed-to-pressed transition, and generates a bleep
				// The waspressed array holds the status of the pad from the last iteration,
				// so comparing to this time around shows transitions
			{
				SetLED(padcount);	  	  				// Turn on LED if key is pressed
				Beep(padcount);
			}
			keypressed |= (1<<padcount);				// set array flag for next time around
		}
		else keypressed &= (1<<padcount)^0xFFFF;		// or clear if pad not pressed
		padcount++;										// next pad
		if (padcount >=NUM_BTTNS)						// if last pad
		{
			padcount= 0;
			if (anykeypressed == 0)
			{
				//TBD
			}
		}
		if (padcount == 0)								// Emergency exit - usually fired off by wet fingers across back of circuit board!
		{												// Test for 4 or more keys pressed at the same time, and do a forced restart if needed.
			for (x =0; x<NUM_BTTNS ;x++)
			{
				testbit  = ( keypressed >> x)& 0x01;

				if (testbit ==1) downkeys++;			// count pressed keys
			}
			if (downkeys >=4)
			{
				// reset_cpu();
			}
			downkeys = 0;
		}
	}
} 														// END OF MAIN


// ***********************************************   Subroutines ******************************************************

void measure_pad(unsigned char channel)				// Performs actual read of key
{

	select_keypad(channel);							// first select the required keypad

	CCP1CON = 0b00101100;							// Turn on oscillator
	T2CON = 0b00000100;
	SelectRef(1);									// select pad as opposed to reference coil
	delay = RX_DELAY;

	while(--delay!=0);								//  stability delay
	both_coils = adc();								// Read ADC - output pad value to RS232 second field

	SelectRef(0);									// switch to reference coil

	delay = REF_DELAY;
	while(--delay!=0);								// stability delay
	ref_coil = adc();								// Read ADC - output reference coil value to RS232 third field

	select_keypad(NUM_BTTNS);						// deselect keys

	CCP1CON = 0b00001100;							// Turn off oscillator
	T2CON = 0b00000000;

	delay = REF_DELAY;
	while(--delay!=0);								// stability delay
	gnd_ref = adc();								// Read ADC - output Vref to RS232 fourth field

	largetemp = (ref_coil - both_coils );			// Subtract reading from background
	largetemp = largetemp << 10;					// Multiply to get sensible scale
	largetemp = largetemp / (gnd_ref - ref_coil );	// Divide by (reference - Vref)

	norm_btn = largetemp;							//  RS232 sixth field = keypad reading

	value = AVERAGE[channel]>>4;

	RAW[channel] = norm_btn<<4;

	threshold = value - (TRIP[channel]>>4);			// effective threshold number

	if (norm_btn < threshold)						// Compare reading to threshold
	{
		anykeypressed = 1;							// set flags if key pressed
		padpressed = 1;
	}
	else
	{
		padpressed = 0;								// update threshold average only if key not pressed.
		if ((largetemp >= 400) & (largetemp <2000))	// screen out any ridiculous values
		{
			AVERAGE[channel] -= value;
			AVERAGE[channel] += largetemp;			// Averaging
		}

	}
}

unsigned int adc(void) 								// Get ADC reading with average of 4 value for smoothing and resolution increase
{
	unsigned int madc1;

	GODONE = 0x1;									// start conversion
	while (GODONE);									// wait while busy

	madc1 = ADRESH<<8 | (ADRESL & ADC_MASK);

	GODONE = 0x1;									// start conversion
	while (GODONE);									// wait while busy
	madc1 += (ADRESH<<8 | (ADRESL & ADC_MASK));

	GODONE = 0x1;									// start conversion
	while (GODONE);									// wait while busy
	madc1 += (ADRESH<<8 | (ADRESL & ADC_MASK));

	GODONE = 0x1;									// start conversion
	while (GODONE);									// wait while busy
	madc1 += (ADRESH<<8 | (ADRESL & ADC_MASK));

	madc1 >>=1;

	return (madc1);

}


void restart_delay(void)
{
	delay = WAIT_100ms;							// 100ms settling delay - used at start to let everything stabilise
	while (--delay!=0);
}



